using System;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Linq;
using Nop.Plugin.Misc.DynamicsCrm.Data;
using Nop.Services.Logging;
using Microsoft.Xrm.Sdk;
using System.ServiceModel.Description;
using Microsoft.Xrm.Sdk.Client;
using Nop.Services.Customers;
using Nop.Core.Domain.Customers;
using Nop.Services.Messages;
using Nop.Plugin.Misc.DynamicsCrm.Models;
using System.Diagnostics;
using Nop.Services.Localization;
using Nop.Core.Domain.Messages;
using System.ServiceModel.Security;
using System.Globalization;
using System.Text;
using Nop.Services.Orders;
using Nop.Core.Domain.Orders;
using Nop.Core.Infrastructure;
using Nop.Plugin.Misc.DynamicsCrm.Catalog.Services;
using System.Web;
using Nop.Core;
using System.Web.Mvc;
using System.Web.Routing;
using Microsoft.Xrm.Sdk.Messages;
using Microsoft.Xrm.Sdk.Metadata;
using Nop.Services.Common;

namespace Nop.Plugin.Misc.DynamicsCrm.Services
{
    public class DynamicsCrmApiService : IDynamicsCrmApiService
    {
        private readonly DynamicsCrmSettings _dynamicsCrmSettings;
        private readonly IDynamicsCrmEventQueueingService _subscriptionEventQueueingService;
        private readonly INewsLetterSubscriptionService _newsLetterSubscriptionService;
        private readonly ICustomerService _customerService;
        private readonly IOrderService _orderService;
        private readonly IMessageTemplateService _messageTemplateService;
        private readonly IQueuedEmailService _queuedEmailService;
        private readonly ILanguageService _languageService;
        private readonly ITokenizer _tokenizer;
        private readonly IEmailAccountService _emailAccountService;
        private readonly IMessageTokenProvider _messageTokenProvider;

        private readonly EmailAccountSettings _emailAccountSettings;

        private readonly ILogger _log;

        public DynamicsCrmApiService(DynamicsCrmSettings dynamicsCrmSettings,
            IDynamicsCrmEventQueueingService subscriptionEventQueueingService,
            INewsLetterSubscriptionService newsLetterSubscriptionService,
            ICustomerService customerService,
            IOrderService orderService,
            IMessageTemplateService messageTemplateService,
            IQueuedEmailService queuedEmailService, ILanguageService languageService,
            ITokenizer tokenizer, IEmailAccountService emailAccountService,
            IMessageTokenProvider messageTokenProvider,
            EmailAccountSettings emailAccountSettings,
            ILogger log)
        {
            _dynamicsCrmSettings = dynamicsCrmSettings;
            _subscriptionEventQueueingService = subscriptionEventQueueingService;
            _newsLetterSubscriptionService = newsLetterSubscriptionService;
            _customerService = customerService;
            _orderService = orderService;

            _messageTemplateService = messageTemplateService;
            _queuedEmailService = queuedEmailService;
            _languageService = languageService;
            _tokenizer = tokenizer;
            _emailAccountService = emailAccountService;
            _messageTokenProvider = messageTokenProvider;

            _emailAccountSettings = emailAccountSettings;

            _log = log;
        }


        IOrganizationService _service;
        ServiceContext _context;

        #region Cached objects
        PriceLevel _PriceLevel;
        public PriceLevel PriceLevel
        {
            get
            {
                if (_PriceLevel == null)
                {
                    _PriceLevel = (from c in _context.PriceLevelSet
                                   where c.Name == _dynamicsCrmSettings.PriceLevel
                                   select c).FirstOrDefault();
                }
                return _PriceLevel;
            }
            set { _PriceLevel = value; }
        }

        UoM _UoM;
        public UoM Uom
        {
            get
            {
                if (_UoM == null)
                {
                    _UoM = (from c in _context.UoMSet
                            where c.Name == _dynamicsCrmSettings.Uom
                            select c).FirstOrDefault();
                }
                return _UoM;
            }
            set { _UoM = value; }
        }


        UoMSchedule _UoMSchedule;
        public UoMSchedule UomSchedule
        {
            get
            {
                if (_UoMSchedule == null)
                {
                    _UoMSchedule = (from c in _context.UoMScheduleSet
                                    where c.Name == _dynamicsCrmSettings.UomSchedule
                                    select c).FirstOrDefault();
                }
                return _UoMSchedule;
            }
            set { _UoMSchedule = value; }
        }

        Dictionary<string, Guid> _Products;
        public Dictionary<string, Guid> Products
        {
            get
            {
                if (_Products == null)
                {
                    _Products = new Dictionary<string, Guid>();
                    var list = (from c in _context.ProductSet
                                select new { c.ProductNumber, c.Id });
                    foreach (var p in list)
                    {
                        try
                        {
                            if (!string.IsNullOrEmpty(p.ProductNumber))
                                _Products.Add(p.ProductNumber, p.Id);
                        }
                        catch (Exception) { }
                    }
                }
                return _Products;
            }
            set { _Products = value; }
        }



        Dictionary<string, int> _ShippingMethods;
        public Dictionary<string, int> ShippingMethods
        {
            get
            {
                if (_ShippingMethods == null)
                {
                    RetrieveEntityRequest retrieveDetails = new RetrieveEntityRequest { EntityFilters = EntityFilters.All, LogicalName = "salesorder" };
                    RetrieveEntityResponse retrieveEntityResponseObj = (RetrieveEntityResponse)_service.Execute(retrieveDetails);
                    EntityMetadata metadata = retrieveEntityResponseObj.EntityMetadata;
                    PicklistAttributeMetadata picklistMetadata = metadata.Attributes.FirstOrDefault(attribute => String.Equals(attribute.LogicalName, "shippingmethodcode", StringComparison.OrdinalIgnoreCase)) as PicklistAttributeMetadata;

                    _ShippingMethods = new Dictionary<string, int>();
                    foreach (var optionMetadata in picklistMetadata.OptionSet.Options)
                    {
                        try
                        {
                            if (!string.IsNullOrEmpty(optionMetadata.Label.LocalizedLabels[0].Label))
                                _ShippingMethods.Add(optionMetadata.Label.LocalizedLabels[0].Label, optionMetadata.Value.Value);
                        }
                        catch (Exception) { }
                    }

                }
                return _ShippingMethods;
            }
            set { _ShippingMethods = value; }
        }

        #endregion

        static bool Synchronizing = false;

        public virtual SyncResult Synchronize()
        {
            var result = new SyncResult();
            if (Synchronizing)
            {
                result.ConnectionError = "Already synchronizing";
            }
            else
            {
                Synchronizing = true;

                try
                {
                    ClientCredentials credentials = new ClientCredentials();

                    credentials.UserName.UserName = _dynamicsCrmSettings.UserName;
                    credentials.UserName.Password = _dynamicsCrmSettings.Password;

                    //Dans le menu parametres du crm, personnalisation, ressources dev
                    Uri organizationUri = new Uri(_dynamicsCrmSettings.OrganizationUri);

                    Uri homeRealmUri = null;
                    OrganizationServiceProxy serviceProxy = new OrganizationServiceProxy(organizationUri, homeRealmUri, credentials, null);
                    serviceProxy.EnableProxyTypes();
                    _service = (IOrganizationService)serviceProxy;
                    _context = new ServiceContext(_service);

                }
                catch (Exception ex)
                {
                    SendErrorMessage(ex.ToStringWithInnerException());
                    result.ConnectionError = ex.ToStringWithInnerException();
                    return result;
                }


                // Get all the queued records for subscription
                var allRecords = _subscriptionEventQueueingService.GetAll();
                int nbSuccess, nbErrors;
                Stopwatch chrono = new Stopwatch();

                //get unique and latest records
                var allRecordsUnique = new List<DynamicsCrmEventQueue>();
                foreach (var item in allRecords
                    .OrderBy(x => x.CreatedOnUtc))
                {
                    var exists = allRecordsUnique
                        .Where(x => x.EntityId == item.EntityId && x.EntityType.Equals(item.EntityType, StringComparison.InvariantCultureIgnoreCase))
                        .FirstOrDefault() != null;
                    if (!exists)
                        allRecordsUnique.Add(item);
                }

                #region contact
                nbSuccess = 0;
                nbErrors = 0;
                chrono.Reset();
                chrono.Start();
                foreach (var contactRecord in allRecordsUnique.Where(x => x.EntityType == "CONTACT").ToList())
                {
                    try
                    {
                        var customer = _customerService.GetCustomerById(contactRecord.EntityId);
                        if (customer != null)
                        {

                            //var contact = (from c in _context.ContactSet
                            //               where c.Id == customer.CustomerGuid
                            //               select c).FirstOrDefault();

                            var contact = (from c in _context.ContactSet
                                           where c.EMailAddress1 == customer.GetValidEmail()
                                           && c.CustomerTypeCode.Value == _dynamicsCrmSettings.CustomerTypeCode
                                           select c).FirstOrDefault();

                            if (contact == null)
                            {
                                contact = CreateContact(customer);
                            }
                            else
                            {
                                SetContactFields(contact, customer);
                                _context.UpdateObject(contact);
                            }
                            _context.SaveChanges();
                        }

                        //delete the queued records
                        _subscriptionEventQueueingService.Delete(contactRecord);
                        nbSuccess++;

                    }
                    catch (MessageSecurityException msex)
                    {
                        nbErrors++;
                        result.ContactErrors.Add(msex.ToStringWithInnerException());
                        _context.ClearChanges();
                        break;
                    }
                    catch (Exception ex)
                    {
                        var link = string.Format("{0}Admin/Customer/Edit/{1}", EngineContext.Current.Resolve<Nop.Core.Domain.StoreInformationSettings>().StoreUrl, contactRecord.EntityId);
                        result.ContactErrors.Add(ex.ToStringWithInnerException(link));
                        nbErrors++;
                        _context.ClearChanges();
                    }
                }
                chrono.Stop();
                result.ContactResult = string.Format("Synchronisation des Contacts termine en {0}s , {1} succs {2} erreurs sur un total de {3}", chrono.Elapsed.Seconds, nbSuccess, nbErrors, nbSuccess + nbErrors);
                if (nbErrors != 0) SendErrorMessage(Concat(result.ContactResult, result.ContactErrors));
                #endregion


                #region Newsletter subscription
                nbSuccess = 0;
                nbErrors = 0;
                chrono.Reset();
                chrono.Start();
                foreach (var newslettersubsRecord in allRecordsUnique.Where(x => x.EntityType == "NEWSLETTERSUBS").ToList())
                {
                    try
                    {
                        var newslettersubs = _newsLetterSubscriptionService.GetNewsLetterSubscriptionById(newslettersubsRecord.EntityId);
                        if (newslettersubs != null)
                        {

                            //var contact = (from c in _context.ContactSet
                            //               where c.Id == customer.CustomerGuid
                            //               select c).FirstOrDefault();

                            var contact = (from c in _context.ContactSet
                                           where c.EMailAddress1 == newslettersubs.Email
                                           && c.CustomerTypeCode.Value == _dynamicsCrmSettings.CustomerTypeCode
                                           select c).FirstOrDefault();

                            if (contact == null)
                            {
                                contact = CreateContact(newslettersubs);
                            }
                            //No need to update contact because we have no more data to set
                            //else
                            //{
                            //    SetContactFields(contact, newslettersubs);
                            //    _context.UpdateObject(contact);
                            //}
                            _context.SaveChanges();
                        }

                        //delete the queued records
                        _subscriptionEventQueueingService.Delete(newslettersubsRecord);
                        nbSuccess++;

                    }
                    catch (MessageSecurityException msex)
                    {
                        nbErrors++;
                        result.NewsLetterSubscriptionErrors.Add(msex.ToStringWithInnerException());
                        _context.ClearChanges();
                        break;
                    }
                    catch (Exception ex)
                    {
                        var link = string.Format("{0}Admin/NewsLetterSubscription/Edit/{1}", EngineContext.Current.Resolve<Nop.Core.Domain.StoreInformationSettings>().StoreUrl, newslettersubsRecord.EntityId);
                        result.NewsLetterSubscriptionErrors.Add(ex.ToStringWithInnerException(link));
                        nbErrors++;
                        _context.ClearChanges();
                    }
                }
                chrono.Stop();
                result.NewsLetterSubscriptionResult = string.Format("Synchronisation des NewsLetterSubscriptions termine en {0}s , {1} succs {2} erreurs sur un total de {3}", chrono.Elapsed.Seconds, nbSuccess, nbErrors, nbSuccess + nbErrors);
                if (nbErrors != 0) SendErrorMessage(Concat(result.NewsLetterSubscriptionResult, result.NewsLetterSubscriptionErrors));
                #endregion


                #region order
                nbSuccess = 0;
                nbErrors = 0;

                Catalog.Services.IDynamicsCrmApiService catalogService;
                //is plugin installed?
                var pluginFinder = EngineContext.Current.Resolve<Nop.Core.Plugins.IPluginFinder>();
                var pluginDescriptor = pluginFinder.GetPluginDescriptorBySystemName("Misc.DynamicsCrm.Catalog", false);
                if (pluginDescriptor == null || !pluginDescriptor.Installed)
                    catalogService = null;
                else
                    catalogService = EngineContext.Current.Resolve<Catalog.Services.IDynamicsCrmApiService>();

                chrono.Reset();
                chrono.Start();
                foreach (var salesOrderRecord in allRecordsUnique.Where(x => x.EntityType == "SALESORDER").ToList())
                {
                    try
                    {
                        var order = _orderService.GetOrderById(salesOrderRecord.EntityId);
                        if (order != null)
                        {
                            //var contact = (from c in _context.ContactSet
                            //               where c.Id == order.Customer.CustomerGuid
                            //               select c).FirstOrDefault();

                            var contact = (from c in _context.ContactSet
                                           where c.EMailAddress1 == order.Customer.GetValidEmail()
                                           && c.CustomerTypeCode.Value == _dynamicsCrmSettings.CustomerTypeCode
                                           select c).FirstOrDefault();

                            if (contact == null) //This case appears when customer is gest
                            {
                                contact = CreateContact(order.Customer);
                                _context.SaveChanges();
                            }

                            var salesOrder = (from c in _context.SalesOrderSet
                                              where c.Id == order.OrderGuid
                                              select c).FirstOrDefault();

                            if (salesOrder == null)
                            {

                                Guid productId;
                                foreach (var orderProductVariant in order.OrderProductVariants)
                                {
                                    if (orderProductVariant.ProductVariant.Sku == null)
                                        throw new ApplicationException(string.Format("Product SKU is empty on ProductVariantId {0}, please fill it", orderProductVariant.ProductVariant.Id));
                                    if (!Products.TryGetValue(orderProductVariant.ProductVariant.Sku, out productId))
                                    {
                                        if (catalogService == null)
                                        {
                                            throw new ApplicationException(string.Format("Product SKU {0} have not equivalent in CRM, please add it. If you need automatic product creation, you can buy complementary plugin on http://store.muniere.fr, and have a look on 'DynamicsCrm bridge for catalog entities'", orderProductVariant.ProductVariant.Sku));
                                        }
                                        else
                                        {
                                            var product = catalogService.CreateProduct(_service, orderProductVariant.ProductVariant, PriceLevel, Uom, UomSchedule);
                                            //Use the local cache
                                            _Products.Add(product.ProductNumber, product.Id);
                                        }
                                    }
                                }

                                salesOrder = CreateSalesOrder(contact, order);


                                //Dans le cas d'un cration, on insre les lignes de commandes sans conditions
                                foreach (var orderProductVariant in order.OrderProductVariants)
                                {
                                    var salesOrderDetail = CreateSalesOrderDetail(salesOrder, orderProductVariant);
                                }
                                _context.SaveChanges();
                            }
                            else
                            {
                                SetOrderFields(salesOrder, order);
                                _context.UpdateObject(salesOrder);
                                _context.SaveChanges();
                            }

                        }

                        //delete the queued records
                        _subscriptionEventQueueingService.Delete(salesOrderRecord);
                        nbSuccess++;

                    }
                    catch (MessageSecurityException msex)
                    {
                        nbErrors++;
                        result.OrderErrors.Add(msex.ToStringWithInnerException());
                        _context.ClearChanges();
                        break;
                    }
                    catch (Exception ex)
                    {
                        var link = string.Format("{0}Admin/Order/Edit/{1}", EngineContext.Current.Resolve<Nop.Core.Domain.StoreInformationSettings>().StoreUrl, salesOrderRecord.EntityId);
                        result.OrderErrors.Add(ex.ToStringWithInnerException(link));
                        _context.ClearChanges();
                        nbErrors++;
                    }
                }
                chrono.Stop();
                result.OrderResult = string.Format("Synchronisation des Orders termine en {0}s , {1} succs {2} erreurs sur un total de {3}", chrono.Elapsed.Seconds, nbSuccess, nbErrors, nbSuccess + nbErrors);
                if (nbErrors != 0) SendErrorMessage(Concat(result.OrderResult, result.OrderErrors));

                #endregion

                Synchronizing = false;
            }


            return result;
        }

        private Contact CreateContact(Customer customer)
        {
            //var contact = new Contact() { Id = customer.CustomerGuid };
            var contact = new Contact() {};
            contact.OverriddenCreatedOn = customer.CreatedOnUtc;
            SetContactFields(contact, customer);
            _context.AddObject(contact);
            return contact;
        }

        private Contact CreateContact(NewsLetterSubscription newsLetterSubscription)
        {
            //var contact = new Contact() { Id = customer.CustomerGuid };
            var contact = new Contact() { };
            contact.OverriddenCreatedOn = newsLetterSubscription.CreatedOnUtc;
            SetContactFields(contact, newsLetterSubscription);
            _context.AddObject(contact);
            return contact;
        }

        private SalesOrder CreateSalesOrder(Contact contact, Order nopOrder)
        {
            var salesOrder = new SalesOrder() { Id = nopOrder.OrderGuid };
            salesOrder.CustomerId = contact.ToEntityReference();
            salesOrder.Name = nopOrder.Id.ToString();
            salesOrder.OrderNumber = nopOrder.Id.ToString();
            salesOrder.OverriddenCreatedOn = nopOrder.CreatedOnUtc;
            salesOrder.PriceLevelId = PriceLevel.ToEntityReference();
            SetOrderFields(salesOrder, nopOrder);
            if (!_context.IsAttached(contact)) _context.Attach(contact);
            _context.AddRelatedObject(contact, new Relationship("contactorders_association"), salesOrder);
            return salesOrder;
        }

        private SalesOrderDetail CreateSalesOrderDetail(SalesOrder salesOrder, OrderProductVariant nopOrderProductVariant)
        {
            var salesOrderDetail = new SalesOrderDetail();
            salesOrderDetail.SalesOrderId = salesOrder.ToEntityReference();

            Guid productId;

            if (nopOrderProductVariant.ProductVariant.Sku == null)
                throw new ApplicationException(string.Format("Product SKU is empty on ProductVariantId {0}, please fill it", nopOrderProductVariant.ProductVariant.Id));
            if (!Products.TryGetValue(nopOrderProductVariant.ProductVariant.Sku, out productId))
                throw new ApplicationException(string.Format("Product SKU {0} have not equivalent in CRM, please add it", nopOrderProductVariant.ProductVariant.Sku));


            salesOrderDetail.ProductId = new EntityReference("product", productId);

            salesOrderDetail.UoMId = Uom.ToEntityReference();
            salesOrderDetail.IsPriceOverridden = true;
            SetOrderDetailFields(salesOrderDetail, nopOrderProductVariant);
            if (!_context.IsAttached(salesOrder)) _context.Attach(salesOrder);
            _context.AddRelatedObject(salesOrder, new Relationship("order_details"), salesOrderDetail);
            return salesOrderDetail;
        }

        private void SetContactFields(Contact contact, Customer nopCustomer)
        {
            Nop.Core.Domain.Common.Address validAddress = null;
            OptionSetValue validAddressType = null;
            if (nopCustomer.BillingAddress != null)
            {
                validAddress = nopCustomer.BillingAddress;
                validAddressType = new OptionSetValue(1);//facturation
            }
            else if (nopCustomer.ShippingAddress != null)
            {
                validAddress = nopCustomer.ShippingAddress;
                validAddressType = new OptionSetValue(2);//envoi
            }
            else if (nopCustomer.Addresses.Count > 0)
            {
                validAddress = nopCustomer.Addresses.Last();
                validAddressType = new OptionSetValue(4);//autre
            }

            string email = nopCustomer.Email;
            if (string.IsNullOrEmpty(email) && validAddress != null)
                email = validAddress.Email;
            contact.EMailAddress1 = email;

            string lastName = nopCustomer.GetAttribute<string>(SystemCustomerAttributeNames.LastName);
            if (string.IsNullOrEmpty(lastName) && validAddress != null)
                lastName = validAddress.LastName;
            contact.LastName = CultureInfo.CurrentCulture.TextInfo.ToTitleCase(lastName.ToLower());

            string firstName = nopCustomer.GetAttribute<string>(SystemCustomerAttributeNames.FirstName);
            if (string.IsNullOrEmpty(firstName) && validAddress != null)
                firstName = validAddress.FirstName;
            contact.FirstName = CultureInfo.CurrentCulture.TextInfo.ToTitleCase(firstName.ToLower());

            if (_dynamicsCrmSettings.CustomerTypeCode != 0)
                contact.CustomerTypeCode = new OptionSetValue(_dynamicsCrmSettings.CustomerTypeCode);

            var gender = nopCustomer.GetAttribute<string>(SystemCustomerAttributeNames.Gender);
            if (gender == "M")
                contact.GenderCode = new OptionSetValue(1);//man
            else if (gender == "F")
                contact.GenderCode = new OptionSetValue(2);//woman
            else
                contact.GenderCode = null;

            contact.BirthDate = nopCustomer.GetAttribute<DateTime?>(SystemCustomerAttributeNames.DateOfBirth);

            if (validAddress != null)
            {
                contact.Address1_AddressTypeCode = validAddressType;
                contact.Address1_Line1 = validAddress.Address1;
                contact.Address1_Line2 = validAddress.Address2;
                contact.Address1_PostalCode = validAddress.ZipPostalCode;
                contact.Address1_City = validAddress.City;
                if (validAddress.StateProvince != null)
                    contact.Address1_StateOrProvince = validAddress.StateProvince.Name;
                if (validAddress.Country != null)
                    contact.Address1_Country = validAddress.Country.Name;
                contact.Address1_Fax = validAddress.FaxNumber;
                contact.Address1_Telephone1 = validAddress.PhoneNumber;
            }

            var newsletter = _newsLetterSubscriptionService.GetNewsLetterSubscriptionByEmail(nopCustomer.Email);
            if (newsletter != null)
            {
                contact.DoNotBulkEMail = !newsletter.Active;
                contact.Attributes["new_newslettersubscriptionguid"] = newsletter.NewsLetterSubscriptionGuid.ToString();
            }
            else //No bulk email
            {
                contact.DoNotBulkEMail = true;
            }

            contact.Attributes["new_backofficeurl"] = string.Format("{0}Admin/Customer/Edit/{1}", EngineContext.Current.Resolve<Nop.Core.Domain.StoreInformationSettings>().StoreUrl, nopCustomer.Id);

            contact.Description = "Ce compte a t import  partir du site web.";
        }

        private void SetContactFields(Contact contact, NewsLetterSubscription newsLetterSubscription)
        {
            contact.EMailAddress1 = newsLetterSubscription.Email;
            contact.LastName = newsLetterSubscription.Email;
            //contact.FirstName = newsLetterSubscription.Email;

            if (_dynamicsCrmSettings.CustomerTypeCode != 0)
                contact.CustomerTypeCode = new OptionSetValue(_dynamicsCrmSettings.CustomerTypeCode);

            contact.DoNotBulkEMail = newsLetterSubscription.Active;
            contact.Attributes["new_newslettersubscriptionguid"] = newsLetterSubscription.NewsLetterSubscriptionGuid.ToString();

            contact.Description = "Ce compte a t import  partir du site web.";
        }

        private void SetOrderFields(SalesOrder salesOrder, Order nopOrder)
        {

            int shippingMethodId;
            if (!ShippingMethods.TryGetValue(nopOrder.ShippingMethod, out shippingMethodId))
            {
                shippingMethodId = 0;
                foreach (var sm in ShippingMethods)
                    if (sm.Value > shippingMethodId) shippingMethodId = sm.Value;
                shippingMethodId = shippingMethodId + 1;

                InsertOptionValueRequest request = new InsertOptionValueRequest();
                request.EntityLogicalName = "salesorder";
                request.AttributeLogicalName = "shippingmethodcode";
                request.Label = new Label(nopOrder.ShippingMethod, 1036);
                request.Value = shippingMethodId;
                
                InsertOptionValueResponse response = (InsertOptionValueResponse)_service.Execute(request);
                _log.Information(string.Format("ShippingMethod '{0}' has been added to list", nopOrder.ShippingMethod));
                //Use the local cache
                _ShippingMethods.Add(nopOrder.ShippingMethod, shippingMethodId);
            }
            salesOrder.ShippingMethodCode = new OptionSetValue(shippingMethodId);

            if (nopOrder.ShippingAddress != null)
            {
                salesOrder.BillTo_ContactName = nopOrder.ShippingAddress.FirstName + " " + nopOrder.ShippingAddress.LastName;
                salesOrder.BillTo_Line1 = nopOrder.ShippingAddress.Address1;
                salesOrder.BillTo_Line2 = nopOrder.ShippingAddress.Address2;
                salesOrder.BillTo_PostalCode = nopOrder.ShippingAddress.ZipPostalCode;
                salesOrder.BillTo_City = nopOrder.ShippingAddress.City;
                if (nopOrder.ShippingAddress.StateProvince != null)
                    salesOrder.BillTo_StateOrProvince = nopOrder.ShippingAddress.StateProvince.Name;
                if (nopOrder.ShippingAddress.Country != null)
                    salesOrder.BillTo_Country = nopOrder.ShippingAddress.Country.Name;
                salesOrder.BillTo_Fax = nopOrder.ShippingAddress.FaxNumber;
                salesOrder.BillTo_Telephone = nopOrder.ShippingAddress.PhoneNumber;
            }

            if (nopOrder.BillingAddress != null)
            {
                //salesOrder.ShipTo_AddressTypeCode = new OptionSetValue(1);//Envoi
                salesOrder.ShipTo_Line1 = nopOrder.BillingAddress.FirstName + " " + nopOrder.BillingAddress.LastName;
                salesOrder.ShipTo_Line1 = nopOrder.BillingAddress.Address1;
                salesOrder.ShipTo_Line2 = nopOrder.BillingAddress.Address2;
                salesOrder.ShipTo_PostalCode = nopOrder.BillingAddress.ZipPostalCode;
                salesOrder.ShipTo_City = nopOrder.BillingAddress.City;
                if (nopOrder.BillingAddress.StateProvince != null)
                    salesOrder.ShipTo_StateOrProvince = nopOrder.BillingAddress.StateProvince.Name;
                if (nopOrder.BillingAddress.Country != null)
                    salesOrder.ShipTo_Country = nopOrder.BillingAddress.Country.Name;
                salesOrder.ShipTo_Fax = nopOrder.BillingAddress.FaxNumber;
                salesOrder.ShipTo_Telephone = nopOrder.BillingAddress.PhoneNumber;
            }

            salesOrder.SubmitDate = nopOrder.CreatedOnUtc.ToLocalTime();

            salesOrder.DiscountAmount = new Money(nopOrder.OrderSubTotalDiscountInclTax);
            salesOrder.FreightAmount = new Money(nopOrder.OrderShippingInclTax);

            salesOrder.Attributes["new_shippingstatus"] = nopOrder.ShippingStatus.ToString();
            var a = nopOrder.Shipments.LastOrDefault();
            if (a != null)
            {
                salesOrder.Attributes["new_trackingnumber"] = a.TrackingNumber;
                if (a.ShippedDateUtc.HasValue)
                    salesOrder.DateFulfilled = a.ShippedDateUtc.Value.ToLocalTime();
            }

            salesOrder.Attributes["new_paymentstatus"] = nopOrder.PaymentStatus.ToString();
            salesOrder.Attributes["new_paymentmethod"] = nopOrder.PaymentMethodSystemName.Replace("Payments.", "");
            if (nopOrder.PaidDateUtc.HasValue)
            {
                salesOrder.Attributes["new_paiddate"] = nopOrder.PaidDateUtc.Value.ToLocalTime();
            }

            salesOrder.Attributes["new_orderstatus"] = nopOrder.OrderStatus.ToString();

            salesOrder.Attributes["new_backofficeurl"] = string.Format("{0}Admin/Order/Edit/{1}", EngineContext.Current.Resolve<Nop.Core.Domain.StoreInformationSettings>().StoreUrl, nopOrder.Id);

        }

        private void SetOrderDetailFields(SalesOrderDetail salesOrderDetail, OrderProductVariant orderProductVariant)
        {

            //salesOrderDetail.ManualDiscountAmount = new Money(Math.Round(orderProductVariant.DiscountAmountInclTax, 2));
            salesOrderDetail.Quantity = orderProductVariant.Quantity;
            salesOrderDetail.PricePerUnit = new Money(Math.Round(orderProductVariant.UnitPriceExclTax, 2));
            //The tax is not a unit tax!
            salesOrderDetail.Tax = new Money(Math.Round(orderProductVariant.PriceInclTax - orderProductVariant.PriceExclTax, 2));
        }

        //private void SetAccountFields(Account account, Customer customer)
        //{
        //    account.EMailAddress1 = customer.Email;
        //    account.Name = customer.GetAttribute<string>(SystemCustomerAttributeNames.Company);
        //    if (string.IsNullOrEmpty(account.Name))
        //    {
        //        account.Name = string.Format("{0} (particulier)", customer.GetFullName());
        //    }          

        //    //// set the account category to 'Preferred Customer'
        //    //AccountCategoryCode = new OptionSetValue(1),
        //    account.Description = "Ce compte a t import  partir du site web.";
        //}

        private string Concat(string message, List<String> messages)
        {
            System.Text.StringBuilder sb = new System.Text.StringBuilder();
            sb.Append(message);
            sb.AppendLine("<br />");
            foreach (var m in messages)
            {
                sb.Append(m);
                sb.AppendLine("<br />");
            }
            return sb.ToString();
        }

        private void SendErrorMessage(string message)
        {
            var tokens = new List<Token>();
            _messageTokenProvider.AddStoreTokens(tokens);

            var emailAccount = _emailAccountService.GetEmailAccountById(_emailAccountSettings.DefaultEmailAccountId);
            if (emailAccount == null)
                emailAccount = _emailAccountService.GetAllEmailAccounts().FirstOrDefault();
            var toEmail = _dynamicsCrmSettings.EmailErrorsTo;

            var subject = "Error on DynamicsCrm plugin";
            var body = "%Store.Name%<br />Error on DynamicsCrm plugin<br />" + message;
            //Replace subject and body tokens 
            var subjectReplaced = _tokenizer.Replace(subject, tokens, false);
            var bodyReplaced = _tokenizer.Replace(body, tokens, true);

            var email = new QueuedEmail()
            {
                Priority = 5,
                From = emailAccount.Email,
                FromName = emailAccount.DisplayName,
                To = toEmail,
                ToName = toEmail,
                CC = string.Empty,
                Bcc = string.Empty,
                Subject = subjectReplaced,
                Body = bodyReplaced,
                CreatedOnUtc = DateTime.UtcNow,
                EmailAccountId = emailAccount.Id
            };

            _queuedEmailService.InsertQueuedEmail(email);
        }

    }

    public static class ExceptionFormatterExtensions
    {
        public static string ToStringWithInnerException(
            this Exception ex, string url = null)
        {
            StringBuilder description = new StringBuilder();
            description.AppendLine(ex.ToString());

            if (ex.InnerException != null)
            {
                description.Append("Inner exception ---> ");
                description.AppendLine(ex.InnerException.ToString());
            }

            if (url != null)
            {
                description.Append("Url of problematic objet---> ");
                description.AppendLine(url);
            }

            description.Append("Stack trace ---> ");
            description.AppendLine(ex.StackTrace);

            return description.ToString();
        }
    }

    public static class CustomerExtensions
    {
        public static string GetValidEmail(this Customer nopCustomer)
        {
            string email = nopCustomer.Email;
            if (string.IsNullOrEmpty(email))
            {
                if (nopCustomer.BillingAddress != null)
                {
                    email = nopCustomer.BillingAddress.Email;
                }
                else if (nopCustomer.ShippingAddress != null)
                {
                    email = nopCustomer.ShippingAddress.Email;
                }
                else if (nopCustomer.Addresses.Count > 0)
                {
                    email = nopCustomer.Addresses.Last().Email;
                }
            }

            return email;
        }
    }

}